from .edge import Edge
from .edge_ref import EdgeRef
from .max_pooling_edge import MaxPoolingEdge
from .convolutional_edge import ConvolutionalEdge
from .deconvolutional_edge import DeconvolutionalEdge

_EDGE_TYPE_LIST = [
    Edge,
    EdgeRef,
    MaxPoolingEdge,
    ConvolutionalEdge,
    DeconvolutionalEdge,
    ]

EDGE_TABLE = {}
for edge_type in _EDGE_TYPE_LIST:
    EDGE_TABLE[edge_type._name] = edge_type



#!/usr/bin/python
#coding=utf-8

"""
This is a setup script for mang.

.. moduleauthor:: Yoonseop Kang <e0engoon@gmail.com>

"""
import os
import subprocess
from setuptools import setup
from setuptools.command.install import install

# Fetch version from git tags, and write to version.py.
# Also, when git is not available (PyPi package), use stored version.py.
# taken from:
# http://blogs.nopcode.org/brainstorm/2013/05/20/pragmatic-python

version_py = os.path.join(os.path.dirname(__file__), "version.py")

try:
    version_git = subprocess.check_output(["git", "describe"]).strip()
except:
    with open(version_py, 'r') as fp:
        version_git = fp.read().strip().split('=')[-1].replace('"', '')

msg = "# do not edit this file!"
with open(version_py, 'w') as fp:
    fp.write(msg + os.linesep + "__version__ = " + version_git)


class MyInstall(install):
    def run(self):
        os.system("cd mang/cudamat; rm *.so; make")
        os.system("cd ../../")
        install.run(self)

CUDAMAT_FILES = [
    "*.cu", "*.cuh", "*.so", "*.txt", "run_on_me_or_pid_quit", "Makefile"]

setup(
    name="mang",
    version=version_git,
    description="Another neural network library for python based on cudamat",

    install_requires=["numpy>=1.6.1", "msgpack-python", "msgpack_numpy"],

    packages=["mang", "mang.node", "mang.edge", "mang.cudamat"],
    package_data={"mang.cudamat": CUDAMAT_FILES, },
    cmdclass={"install": MyInstall},

    author="Yoonseop Kang",
    author_email="e0engoon@gmail.com",
    url="http://github.com/e0en/mang",
    keywords="neural network, artificial intelligence, machine learning",
    zip_safe=False)



#!/usr/bin/python
#coding=utf-8

"""
mang.cpu_conv
~~~~~~~~~~~~~

CPU-version of convolution routines of cudamat-conv.

.. moduleauthor:: Yoonseop Kang <e0engoon@gmail.com>

"""
import os
import ctypes as ct

import numpy as np

current_dir = os.path.realpath(__file__)
dll_path = os.path.join(current_dir, "libcpuconv.so")
_conv = ct.CDLL(dll_path)


def conv_up(x, filters, n_channel, stride):
    """Apply convolution on x using filters.

    Arguments:
        x: input images, (n_image, x_size ** 2 * n_channel)
        filters: convolution filters,
                 (n_filter, filter_size ** 2 * n_channel)
        n_channel: number of colors in images
        n_filter: number of filters
        stride: distance between convolution locations in pixels

    Returns:
        o: convoluted images

    Raises:
        Nothing.

    """
    n_image = x.shape[0]
    x_size = int(np.sqrt(x.shape[1] / n_channel))
    n_filter = filters.shape[0]
    filter_size = int(np.sqrt(filters.shape[1] / n_channel))
    o_size = (x_size - filter_size) / stride + 1

    """
    x = np.array(x, dtype=np.float32)
    o = np.zeros((n_image, o_size ** 2 * n_filter), dtype=np.float32)
    _conv.conv_up(x.ctypes.data_as(ct.POINTER),
                  o.ctypes.data_as(ct.POINTER),
                  x_size, n_filter, filter_size, o_size)
    return o
    """

    for i_image in xrange(n_image):
        img_x = x[i_image].reshape((x_size, x_size, n_channel), order="F")
        img_o = o[i_image].reshape((o_size, o_size, n_filter), order="F")
        for i_filter in xrange(n_filter):
            img_filter = \
                filters[i_filter].reshape((filter_size, filter_size,
                                          n_channel), order="F")
            for i_x in xrange(o_size):
                for i_y in xrange(o_size):
                    i1 = i_x * stride
                    i2 = i1 + filter_size
                    j1 = i_y * stride
                    j2 = j1 + filter_size
                    img_o[i_x, i_y, i_filter] = \
                        (img_x[i1:i2, j1:j2, :] * img_filter).sum()
        o[i_image] = img_o.reshape((o_size * o_size * n_filter), order="F")
    return o


def max_pool(x, n_channel, ratio, stride):
    """Apply convolution on x using filters.

    Arguments:
        x: input images, (n_image, x_size ** 2 * n_channel)
        n_channel: number of colors in images
        ratio: the width of pooling area
        stride: distance between convolution locations in pixels

    Returns:
        o: subsampled images

    Raises:
        Nothing.

    """

    n_image = x.shape[0]
    x_size = int(np.sqrt(x.shape[1] / n_channel))
    o_size = (x_size - ratio) / stride + 1
    o = np.zeros((n_image, o_size ** 2 * n_channel))

    """
    x = np.array(x, dtype=np.float32)
    o = np.zeros((n_image, o_size ** 2 * n_channel), dtype=np.float32)
    _conv.max_pool(x.ctypes.data_as(ct.POINTER),
                  o.ctypes.data_as(ct.POINTER),
                  x_size, n_channel, ratio, stride)
    return o
    """

    for i_image in xrange(n_image):
        img_x = x[i_image].reshape((x_size, x_size, n_channel), order="F")
        img_o = o[i_image].reshape((o_size, o_size, n_channel))
        for i_channel in xrange(n_channel):
            for i_x in xrange(o_size):
                for i_y in xrange(o_size):
                    i1 = i_x * stride
                    i2 = i1 + ratio
                    j1 = i_y * stride
                    j2 = j1 + ratio
                    tmp = img_x[i1:i2, j1:j2, i_channel].ravel()
                    img_o[i_x, i_y, i_channel] = tmp.max()
        o[i_image] = img_o.reshape((n_channel * o_size * o_size, ), order="F")
    return o


def response_normalization(x, n_channel, size, add_scale, pow_scale):
    """Run response normalization on inputs.
    o = x * (1 + add_scale * cov)^(-pow_scale)
    cov = sum_(region) x^2
    region size = (size, size)

    Arguments:
        x: input images, (n_image, x_size ** 2 * n_channel)
        n_channel: number of colors in images
        size: the width of response normalization area

    Returns:
        o: response-normalized images

    Raises:
        Nothing.

    """

    n_image = x.shape[0]
    x_size = int(np.sqrt(x.shape[1] / n_channel))
    o = np.array(x, order="F")
    p = np.array(x, order="F")

    """
    x = np.array(x, dtype=np.float32)
    o = np.zeros((n_image, x_size ** 2 * n_channel), dtype=np.float32)
    _conv.response_normalization(x.ctypes.data_as(ct.POINTER),
                                 o.ctypes.data_as(ct.POINTER),
                                 x_size, n_channel, ratio, stride)
    return o
    """


    for i_image in xrange(n_image):
        img_x = x[i_image].reshape((x_size, x_size, n_channel), order="F")
        img_o = o[i_image].reshape((x_size, x_size, n_channel), order="F")
        img_p = p[i_image].reshape((x_size, x_size, n_channel), order="F")
        for i_channel in xrange(n_channel):
            for i_x in xrange(x_size):
                for i_y in xrange(x_size):
                    i0 = i_x - size / 2
                    j0 = i_y - size / 2
                    i1 = max(0, i0)
                    j1 = max(0, j0)
                    i2 = min(x_size, i0 + size)
                    j2 = min(x_size, j0 + size)
                    prod = img_x[i1:i2, j1:j2, :]
                    prod = 1. + add_scale * (prod * prod).sum(0).sum(0)
                    img_p[i_x, i_y, :] = prod
                    img_o[i_x, i_y, :] = \
                        img_x[i_x, i_y, :] * np.power(prod, -pow_scale)

        o[i_image] = img_o.reshape((n_channel * x_size * x_size, ), order="F")
        p[i_image] = img_p.reshape((n_channel * x_size * x_size, ), order="F")
    return (o, p)



import os
import ctypes as ct
import math
import pdb


current_dir = os.path.dirname(os.path.realpath(__file__))
dll_path = os.path.join(current_dir, 'libcudamat_conv.so')
_ConvNet = ct.cdll.LoadLibrary(dll_path)

def convUp(images, filters, targets, numModulesX, paddingStart, moduleStride, numImgColors, numGroups=1):
  """
  images - (n_images, img_w**2 * n_chans)
  filters - (n_filters, filter_w**2 * n_chans)
  targets - (n_images, n_locs**2 * n_filters)
  numModulesX - Number of filter locations along an axis. = n_locs
  paddingStart - Set to k for a k-pixel border of zeros. Usually set to 0.
  moduleStride - stride to move the filters by. 
  numImgColors - n_chans
  """
  numImages = images.shape[0]
  numFilters = filters.shape[0]

  assert targets.shape == (numImages, numFilters * numModulesX * numModulesX), '%s %d %d-%d-%d' % (targets.shape.__str__(), numImages, numFilters, numModulesX, numModulesX)

  _ConvNet.convUp(images.p_mat, filters.p_mat, targets.p_mat, numModulesX,
                  -paddingStart, moduleStride, numImgColors, numGroups)

def convDown(hidSums, filters, targets, numModulesX, paddingStart, moduleStride, filterSizeX, imSizeX, numImgColors):
  """
  hidSums - (n_images, n_locs**2 * n_filters)
  filters - (n_filters, filter_w**2 * n_chans)
  targets - (n_images, img_w**2 * n_chans)
  """
  numGroups = 1
  numFilters = filters.shape[0]
  numImages = hidSums.shape[0] 
  numModules = numModulesX**2 

  assert paddingStart >= 0
  assert targets.shape == (numImages, numImgColors * imSizeX * imSizeX)

  _ConvNet.convDown(hidSums.p_mat, filters.p_mat, targets.p_mat, imSizeX,
                    -paddingStart, moduleStride, numImgColors, numGroups)


def convOutp(images, hidSums, targets, numModulesX, paddingStart, filterSizeX, moduleStride, numImgColors):
  """
  images - (n_images, img_w**2 * n_chans)
  hidSums - (n_images, n_locs**2 * n_filters)
  targets - (n_filters, filter_w**2 * n_chans)
  """
  numGroups = 1
  partialSum = 0
  numImages = images.shape[0]
  numFilters = hidSums.shape[1] / (numModulesX**2)

  assert targets.shape == (numFilters, numImgColors * filterSizeX * filterSizeX), '%s %d %d-%d-%d' % (targets.shape.__str__(), numFilters, numImgColors, filterSizeX, filterSizeX)
  _ConvNet.convOutp(images.p_mat, hidSums.p_mat, targets.p_mat, numModulesX, filterSizeX, -paddingStart, moduleStride, numImgColors, 1, 0)

def localUp(images, filters, targets, numModulesX, paddingStart, moduleStride, numImgColors, numGroups=1):
  """
  images - (n_images, img_w**2 * n_chans)
  filters - (n_filters, filter_w**2 * n_chans)
  targets - (n_images, n_locs**2 * n_filters)
  numModulesX - Number of filter locations along an axis. = n_locs
  paddingStart - Set to k for a k-pixel border of zeros. Usually set to 0.
  moduleStride - stride to move the filters by. 
  numImgColors - n_chans
  """
  numImages = images.shape[0]
  numFilters = filters.shape[0]

  assert targets.shape == (numImages, numFilters * numModulesX * numModulesX), '%s %d %d-%d-%d' % (targets.shape.__str__(), numImages, numFilters, numModulesX, numModulesX)

  _ConvNet.localUp(images.p_mat, filters.p_mat, targets.p_mat,
           numModulesX, -paddingStart, moduleStride, numImgColors, numGroups)

def localDown(hidSums, filters, targets, numModulesX, paddingStart, moduleStride, filterSizeX, imSizeX, numImgColors):
  """
  hidSums - (n_images, n_locs**2 * n_filters)
  filters - (n_filters, filter_w**2 * n_chans)
  targets - (n_images, img_w**2 * n_chans)
  """
  numGroups = 1
  numFilters = filters.shape[0]
  numImages = hidSums.shape[0] 
  numModules = numModulesX**2 

  assert paddingStart >= 0
  assert targets.shape == (numImages, numImgColors * imSizeX * imSizeX)

  _ConvNet.localDown(hidSums.p_mat, filters.p_mat, targets.p_mat,
            imSizeX, -paddingStart, moduleStride, numImgColors, numGroups)


def localOutp(images, hidSums, targets, numModulesX, paddingStart, filterSizeX, moduleStride, numImgColors):
  """
  images - (n_images, img_w**2 * n_chans)
  hidSums - (n_images, n_locs**2 * n_filters)
  targets - (n_filters, filter_w**2 * n_chans)
  """
  numGroups = 1
  partialSum = 0
  numImages = images.shape[0]
  numFilters = hidSums.shape[1] / (numModulesX**2)

  assert targets.shape == (numFilters, numModulesX**2 * numImgColors * filterSizeX**2), '%s %d %d-%d-%d' % (targets.shape.__str__(), numFilters, numImgColors, filterSizeX, filterSizeX)
  _ConvNet.localOutp(images.p_mat, hidSums.p_mat, targets.p_mat,
            numModulesX, filterSizeX, -paddingStart, moduleStride, numImgColors, numGroups, partialSum)



def MaxPool(images, targets, numChannels, subsX, startX, strideX, outputsX):
  """
  images - (n_images, img_w**2 * n_chans)
  numChannels - number of filter/color channels
  subsX - width of pooling area
  startX - pixel where pooling starts
  strideX - stride
  outputsX - number of pooling sites
  """
  numImages = images.shape[0]

  assert targets.shape == (numImages, numChannels * outputsX * outputsX)
  
  _ConvNet.MaxPool(images.p_mat, targets.p_mat,
           numChannels, subsX, startX, strideX, outputsX)

def ProbMaxPool(images, rnd, targets, numChannels, subsX, startX, strideX, outputsX):
  """
  images - (n_images, img_w**2 * n_chans)
  rnd - (n_images, img_w**2 * n_chans)
  numChannels - number of filter/color channels
  subsX - width of pooling area
  startX - pixel where pooling starts
  strideX - stride
  outputsX - number of pooling sites
  """
  numImages = images.shape[0]

  assert targets.shape == (numImages, numChannels * outputsX * outputsX)
  assert rnd.shape == images.shape

  _ConvNet.ProbMaxPool(images.p_mat, rnd.p_mat, targets.p_mat,
           numChannels, subsX, startX, strideX, outputsX)


def MaxPoolUndo(images, targets, grad, maxes,
        subsX, startX, strideX, outputsX):
  """
  images - (n_images, img_w**2 * n_chans)
  grad - (n_images, outputsX**2 * n_chans) cudamat of deltas/gradients of loss wrt layer outputs.
  maxes - (n_images, outputsX**2 * n_chans) cudamat of layer outputs.
  subsX - width of pooling area
  startX - pixel where pooling starts
  strideX - stride
  outputsX - number of pooling sites
  """
  assert targets.shape == images.shape

  _ConvNet.MaxPoolUndo(images.p_mat, grad.p_mat, maxes.p_mat, targets.p_mat,
             subsX, startX, strideX, outputsX)

def ResponseNorm(images, denoms, targets, numChannels, sizeX, addScale, powScale):
  assert targets.shape == images.shape
  assert targets.shape == denoms.shape
  num_images = images.shape[0]
  numpixels = images.shape[1] / numChannels
  imgsize = int(math.sqrt(numpixels))
  #assert images.shape[1] == numChannels * numpixels
  #assert imgsize * imgsize == numpixels
  #pdb.setrace()
  _ConvNet.ResponseNorm(images.p_mat, denoms.p_mat, targets.p_mat,
             numChannels, sizeX, ct.c_float(addScale),
             ct.c_float(powScale))

def ResponseNormUndo(outGrad, denoms, inGrad, acts, targets, numChannels, sizeX,
           addScale, powScale):
  assert targets.shape == outGrad.shape
  assert targets.shape == denoms.shape
  assert targets.shape == inGrad.shape
  assert targets.shape == acts.shape
  _ConvNet.ResponseNormUndo(outGrad.p_mat, denoms.p_mat, inGrad.p_mat,
               acts.p_mat, targets.p_mat, numChannels, sizeX,
               ct.c_float(addScale), ct.c_float(powScale))
